/* PC-Link is client - server communication system      */
/* designed for transfering data between C64 and PC     */
/* It is part of the IDE64 project http://ide64.come.to */

/* Original design created by Jan Hlavacek in 1998
   Reworked design and DOS/Win version by Tomas Pribyl 1999-2001
   Open source C Linux version in 2002 by Josef Soucek

   Copyright (C) 1998 Jan Hlavacek
   Copyright (C) 1999-2001 Tomas Pribyl <dundera@volny.cz>
   Copyright (C) 2002 Josef Soucek <josef.soucek@ct.cz>

   You can use copy and freely distribute this software in source
   or binary executable format.
   If you want to contribute the development, please contact 
   Josef Soucek <josef.soucek@ct.cz>
   There is no warranty.

   Mar 11 2002 Beta release.
   Mar 19 2002 First release.
   More information read README file in the source tree
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pclink.h"
#include "files.h"
#include "crc8.h"
#include "ports.h"

// #define DEBUG

/* Prototypes */
static int receivepacket(PACKET *);
static int sendpacket(PACKET *);
static unsigned char *getstringfrompacket(PACKET *);
static unsigned char petascii2ascii(unsigned char);
//static unsigned char ascii2petascii(unsigned char);
static int sendstatus(PACKET *, unsigned char);
static int readdir(PACKET *, char *, int, int);

/* Variables */
static unsigned char tmpstring[256];

int main(int argv,char *argc[])
{
  int commstate=stwaitforaction;          /* communication state machine*/
  int command;
  int dirpointer, dirend;
  unsigned char inbyte;
  PACKET inpacket, outpacket;
  char *dirbuff;
  FILE *fr,*fw;

  printf("PC-Link Server V0.1.7\n");
  printf("Information about the IDE64 project http://ide64.come.to\n");
  printf("There is no warranty\n\n");

  if (!initport(PORT,1)) exit(1);

  if (argv==1)
    changeroot(NULL);                      /* Changes path "/" to current dir or root dir */
  if (argv==2)
    changeroot(argc[1]);
  if (argv>2)
  {
    printf("usage: pclink [path]\n");
    exit(1);
  }

  while(true)
  {
    switch(commstate)
    {

      case stwaitforaction:
        printf("Waiting for command from C64...\n");
        timeout=0;                                         /* timeout off*/
        inbyte=getbyte();                                  /* receive one byte*/
	#ifdef DEBUG
          printf("0x%02x ",inbyte);
	#endif
        sendbyte (inbyte ^ 0xff);                          /* send echo*/
        if (inbyte==cactivatecmd)  commstate=stdetectcmd;
        else printf("Activate command not detected...\n");
	break;

      case stdetectcmd:
	printf("Action...\n");
	timeout=TIMEOUT;
	if(receivepacket(&inpacket))
        {
		printf("OK\n");
          switch ((int)inpacket.packettype)
          { 
	    case csendfilecmd:    commstate=stsendfile0; break;         /* sending file*/
	    case cgetfilecmd:     commstate=stgetfile0; break;          /* receiving file*/
	    case ccommandcmd:     commstate=stmakecommand0; break;      /* make command*/
	  }
	}
        else
        {
		printf("Err\n");
          if (inpacket.headererr)  printf("Header CRC error!!\n");
          if (timeouterr)  printf("TimeOut error\n");
          commstate=stwaitforaction;   /* no valid Packet type*/
        }
        break;

      case stsendfile0:
        #ifdef DEBUG
          printf("SF0\n");
        #endif
	getstringfrompacket(&inpacket);
        command=tmpstring[1];	
        if (command=='$') commstate=stsenddir1;
        else commstate=stsendfile1;
        break;

      case stsendfile1:
        #ifdef DEBUG
          printf("SF1\n");
        #endif
        printf("Sending file %s\n",&tmpstring[1]);
        fr=openfile(&tmpstring[1]);
        if (fr==NULL)
        {
          printf("Open file error...\n");
          sendstatus(&outpacket,copenerrst);
          commstate=stwaitforaction;
        }
        else
        {
          sendstatus(&outpacket,cokst);
          commstate=stsendfile2;
        }
        break;

      case stsendfile2:
        #ifdef DEBUG
          printf("SF2\n");
        #endif
        timeout=10;                      /* timeout off*/
        receivepacket(&inpacket);       /* get status packet from MASTER*/
        if (! timeouterr && ! inpacket.headererr)
        {
          switch (inpacket.packettype)
          {
            case cokst:
              commstate=stsendfile3;
              break;                    /* status ok, go ahead*/

            case ccommerrst:
              printf("Repeating Packet..");
              sendpacket(&outpacket);       /* repeat last packet*/
              break;

// In original Tomas version ceofst is not present here
//
//            case ceofst:
//              commstate=stwaitforaction;
//              fclose(fr);
//              break;               /* close communication before last packet*/
              
            case ceotst:
              commstate=stwaitforaction;  /* terminate communication*/
              fclose(fr);
              #ifdef DEBUG
                printf("Communication terminated...\n");
              #endif
              break;
          }
        }
        else
         if (inpacket.headercrc==cactivatecmd)
         {
           fclose(fr);
           commstate=stwaitforaction;     /* received new activate command*/
         }
         else
           printf("Looping SF2...\n");
        break;

      case stsendfile3:
        #ifdef DEBUG
          printf("SF3\n");
        #endif
        timeout=TIMEOUT;

        outpacket.packetlength=fread(&outpacket.packetdata,1,0xff,fr); /* read part into packet */
        if (outpacket.packetlength==0)
        {
          sendstatus(&outpacket,ceofst);           /* send status packet EOF*/
          commstate=stwaitforaction;
          fclose(fr);
        }
        else
        {
          outpacket.packettype=cdatatype;
          sendpacket(&outpacket);                   /* send DATA PACKET*/
          commstate=stsendfile2;
        }
        break;

      case stsenddir1:
        #ifdef DEBUG
          printf("SD1\n");
        #endif
        if ((dirbuff=malloc(DIRBUFF_L))==NULL)
        { 
          printf("Error! Can't allocate buffer\n");
          exit(1);
        }
//        dirend=preparedirectory(dirbuff,DIRBUFF_L);
        dirend=preparedirectory(dirbuff);
        dirpointer=0;
        printf("Sending directory listing...[%d]\n",dirend);
        sendstatus(&outpacket,cokst);
        commstate=stsenddir2;
        break;

      case stsenddir2:
        #ifdef DEBUG
          printf("SD2\n");
        #endif
        timeout=10;                      /* timeout off*/
        receivepacket(&inpacket);       /* get status packet from MASTER*/
        if (! timeouterr && ! inpacket.headererr)
        {
          switch (inpacket.packettype)
          {
            case cokst:
              commstate=stsenddir3;
              break;                    /* status ok, go ahead*/

            case ccommerrst:
              printf("Repeating Packet..");
              sendpacket(&outpacket);       /* repeat last packet*/
              break;

            case ceofst:
              commstate=stwaitforaction;
              free(dirbuff);
              break;               /* close communication before last packet*/
              
            case ceotst:
              commstate=stwaitforaction;  /* terminate communication*/
              free(dirbuff);
              #ifdef DEBUG
                printf("Communication terminated...\n");
              #endif
              break;
          }
        }
        else
         if (inpacket.headercrc==cactivatecmd)
         {
           free(dirbuff);
           commstate=stwaitforaction;     /* received new activate command*/
         }
         else
           printf("Looping SD2...\n");
        break;

      case stsenddir3:
        #ifdef DEBUG
          printf("SD3\n");
        #endif
        timeout=TIMEOUT;

        outpacket.packetlength=readdir(&outpacket,dirbuff,dirpointer,dirend); /* read part into packet */
        dirpointer += outpacket.packetlength;

        if (outpacket.packetlength==0)
        {
          sendstatus(&outpacket,ceofst);           /* send status packet EOF*/
          commstate=stwaitforaction;
          free(dirbuff);
        }
        else
        {
          outpacket.packettype=cdatatype;
          sendpacket(&outpacket);                   /* send DATA PACKET*/
          commstate=stsenddir2;
        }
        break;

      case stgetfile0:
        #ifdef DEBUG
        printf("GF0\n");
        #endif
        getstringfrompacket(&inpacket);
        commstate=stgetfile1;
        break;

      case stgetfile1:
        #ifdef DEBUG
        printf("GF1\n");
        #endif
        printf("Getting file %s\n",&tmpstring[1]);
        fw=createfile(&tmpstring[1]);             /* open file for writing*/
        if (fw==NULL)
        {
          printf("Create file error...\n");
          sendstatus(&outpacket,copenerrst);
          commstate=stwaitforaction;
        }
        else
        {
          sendstatus(&outpacket,cokst);
          commstate=stgetfile2;
        }
        break;

      case stgetfile2:
        #ifdef DEBUG
          printf("GF2\n");
        #endif
        timeout=10;                      /* timeout off*/
        receivepacket(&inpacket);
        if (! timeouterr && ! inpacket.headererr)
        {
          switch (inpacket.packettype)
          {
            case cdatatype:

              if ((fwrite(&inpacket.packetdata,1,inpacket.packetlength,fw))!=inpacket.packetlength)
              {
                printf("Error writing the file..\n");
                fclose(fw);
                commstate=stwaitforaction;
              }
              else
                commstate=stgetfile3;
              break;

            case ceofst:
              commstate=stwaitforaction;
              fclose(fw);
              break;
              
            case ceotst:
              commstate=stwaitforaction;  /* terminate communication*/
              fclose(fw);
              #ifdef DEBUG
                printf("Communication terminated...\n");
              #endif
              break;
          }
        }
        else
         if (inpacket.headercrc==cactivatecmd)
         {
           fclose(fw);
           commstate=stwaitforaction;     /* received new activate command*/
         }
         else
           printf("Looping GF2...\n");
        break;

      case stgetfile3:
        #ifdef DEBUG
          printf("GF3\n");
        #endif
        sendstatus(&outpacket,cokst);       /* send status packet */
        commstate=stgetfile2;
        break;

      case stmakecommand0:
        #ifdef DEBUG
          printf("MC0\n");
        #endif
	getstringfrompacket(&inpacket);
        command=tmpstring[1];	
//        printf("tmp0 %x tmp1 %c tmp2 %x %c\n",tmpstring[0],tmpstring[1],tmpstring[2],tmpstring[2]);
        if ((tmpstring[0]<2) || (tmpstring[2]!=':'))
        {
          sendstatus(&outpacket,cunknownerrst);
          commstate=stwaitforaction;
          printf("Command packet err... %c\n",command);
        }
        else
          commstate=stmakecommand1;
        break;

      case stmakecommand1:
        #ifdef DEBUG
          printf("MC1\n");
        #endif
        switch (command)
        {
          case 's':
            printf("Deleting file: %s\n",&tmpstring[3]);
            if (deletefile(&tmpstring[3]))
              sendstatus(&outpacket,cokst);
            else
            {
              printf("Failed...\n");
              sendstatus(&outpacket,cdeleteerrst);
            }
            break;

          case 'c':
            printf("Changing dir path to: %s\n",&tmpstring[3]);
            if (setnewdirpath(&tmpstring[3]))
              sendstatus(&outpacket,cokst);
            else
            {
              printf("Failed...\n");
              sendstatus(&outpacket,cunknownerrst);
            }
            break;

          case 'm':
            printf("Making directory: %s\n",&tmpstring[3]);
            if (makenewdirectory(&tmpstring[3]))
              sendstatus(&outpacket,cokst);
            else
            {
              printf("Failed...\n");
              sendstatus(&outpacket,cunknownerrst);
            }
            break;

          case 'r':
            if(strchr(&tmpstring[3],'='))
            {
              printf("Renaming file filed. No \"=\" character...\n");
              sendstatus(&outpacket,cunknownerrst);
            }
            else
            {
              if (renamefile(&tmpstring[3]))
                sendstatus(&outpacket,cokst);
              else
              {
                printf("Failed...\n");
                sendstatus(&outpacket,crenameerrst);
              }
            }
            break;

          default:
            printf("Unknown command %c error...\n",command);
            sendstatus(&outpacket,cunknownerrst);
            break;
        }
        commstate=stwaitforaction;
        break;


      default:
        printf("Unknown commstate...\n");
        commstate=stwaitforaction;
        break;
    }
  }
}

static int receivepacket(PACKET *with)
{
  int a;
  #ifdef DEBUG
    printf("receivepacket()\n");
  #endif
  with->headererr=false;

  with->packettype=getbyte();         /* get packet type*/
  if (timeouterr)  return false;
  with->packetlength=getbyte();       /* get Packet Length*/
  if (timeouterr)  return false;
  with->headercrc=getbyte();          /* get Header CRC*/
  if (timeouterr)  return false;
  #ifdef DEBUG
    printf("RECEIVED Packet type:0x%02x; Packet length:0x%02x;\n",with->packettype,with->packetlength);
  #endif

  setresidue(0);
  makecrc(with->packettype);
  makecrc(with->packetlength);       /* Make Header checksum*/
  makecrc(with->headercrc);
  if (getresidue() != 0)             /* test HEADER checksum*/
  {
    with->headererr=true;
    return false;
  }
  if (with->packetlength>0)
  {
    setresidue(0);                   /* initialize CRC8 residue*/
    for(a=0; a<=with->packetlength; a++)
    {
      with->packetdata[a]=getbyte();
      if (timeouterr)  return false;
      makecrc(with->packetdata[a]);  /* make data checksum*/
    }

/* There is no check of the counted and received CRC ??? */

  with->datacrc=getresidue();        /* checksum*/
  }                                                           
  return true;
}

static int sendpacket(PACKET *with)
{
  int i;
  #ifdef DEBUG
    printf("sendpacket()\n");
  #endif
  
  setresidue(0);
  makecrc(with->packettype);
  makecrc(with->packetlength);
  with->headercrc=getresidue();      /* Get HeaderCRC*/
                    
  if (with->packetlength>0)
  {
    setresidue(0);                   /* init data crc*/
    for(i=0; i<=with->packetlength-1; i++) makecrc(with->packetdata[i]);
    with->datacrc=getresidue();      /* Get DataCRC*/
  }

  sendbyte(with->packettype);        /* send PacketType*/
  if (timeouterr)  return false;
  sendbyte(with->packetlength);      /* send PacketLength*/    
  if (timeouterr)  return false;
  sendbyte(with->headercrc);
  if (timeouterr)  return false;    /* send HeaderCRC*/

  if (with->packetlength>0)
  {
    for(i=0; i<=with->packetlength-1; i++)
    {
      sendbyte(with->packetdata[i]); /* send Data*/
      if (timeouterr)  return false; 
    }
  sendbyte(with->datacrc);           /* send DataCRC*/
  }

  #ifdef DEBUG
    printf("SENT Packet type:0x%02x; Packet length:0x%02x;\n",with->packettype,with->packetlength);
  #endif
  return true;
}

static unsigned char *getstringfrompacket(PACKET *with)
{
  int a;

  tmpstring[0]=with->packetlength;
  for(a=0; a<with->packetlength; a++)
    tmpstring[a+1]=petascii2ascii(with->packetdata[a]);
  tmpstring[a+1]='\0';                   /* String terminating */
  return tmpstring;
}

static unsigned char petascii2ascii(unsigned char data)
{
  if ((data>=0xc1) && (data<=0xda))  return (data & 0x7f);
  if ((data>=0x41) && (data<=0x5a))  return (data + 0x20);
  return data;
}
/*
unsigned char ascii2petascii(unsigned char data)
{
  if ((data>=0x61) && (data<=0x7a))  data=(data & 0xdf);
  if ((data>=0x41) && (data<=0x5a))  data=(data | 0x80);
  return data;
}
*/

static int sendstatus(PACKET *with,unsigned char statusnumber)
{
  with->packettype=statusnumber; 
  with->packetlength=0;
  return(sendpacket(with));
}

/* Read part of dir into packet */

static int readdir(PACKET *with, char *dirbuff, int dirpointer, int dirend)
{
  int i;

  printf("READDIR Dirpointer: %d Dirend: %d\n",dirpointer, dirend);

  for(i=0; ((dirpointer<dirend) && (i!=255)); i++)
    with->packetdata[i]=dirbuff[dirpointer++]; 

  printf("READDIR Dirpointer: %d Dirend: %d Paclen:%d\n",dirpointer, dirend, i);
  return i;
}

